"""
Defines ImageButton class, a special Button that includes an image in a frame
and eventually some text next to it.
"""

import os
from typing import Callable, Sequence

import pygame

from .button import Button
from ..configuration import default_sprites
from ..constants import WHITE, MIDNIGHT_BLUE, IMAGE_BUTTON_SIZE
from ..types import Position, Margin


class ImageButton(Button):
    """
    This component is a Button having an image displayed
    at the left of its content and that is considered as part of the button.
    The image is displayed in a dedicated square frame.

    Keyword arguments:
    callback -- the reference to the function that should be call after a click
    size -- the size of the button following the format "(width, height)"
    image_path -- the relative path to the image that should be displayed on the left
    title -- the text that should be displayed at the center of the element
    position -- the position of the element on the screen
    background_path -- the path to the image corresponding to the sprite of the element
    frame_background_path -- the path to the background for the frame
    frame_background_hover_path -- the path to the background for the frame when the mouse is over
    margin -- a tuple containing the margins of the box,
    should be in the form "(top_margin, right_margin, bottom_margin, left_margin)"
    disabled -- a boolean indicating if it is not possible to interact with the button
    font -- the font that should be used to render the text content
    text_color -- the color of the text content
    font_hover -- the font that should be used to render the text content when the mouse is over
    the button
    text_hover_color -- the color of the text content when the mouse is over the button
    complementary_text_lines -- the other text lines that should be displayed in addition of
    the title
    """

    def __init__(
        self,
        image_path: str = None,
        callback: Callable = lambda: None,
        size: tuple[int, int] = IMAGE_BUTTON_SIZE,
        title: str = "",
        position: Position = pygame.Vector2(0, 0),
        background_path: str = None,
        frame_background_path: str = None,
        frame_background_hover_path: str = None,
        margin: Margin = (0, 0, 0, 0),
        disabled: bool = False,
        font: pygame.font.Font = None,
        text_color: tuple[int, int, int] = WHITE,
        font_hover: pygame.font.Font = None,
        text_hover_color: tuple[int, int, int] = MIDNIGHT_BLUE,
        complementary_text_lines: Sequence[str] = None,
    ) -> None:
        super().__init__(
            callback,
            size,
            title,
            position,
            background_path,
            background_path,
            False,
            margin,
            disabled,
            font,
            text_color,
            font_hover,
            text_hover_color,
            complementary_text_lines,
        )

        padding: int = size[1] // 10
        frame_position: Position = pygame.Vector2(padding, padding)
        frame_size: tuple[int, int] = (
            self.size[1] - padding * 2,
            self.size[1] - padding * 2,
        )

        frame_background_path = (
            os.path.abspath(frame_background_path)
            if frame_background_path
            else default_sprites["button_background"]["inactive"]
        )
        raw_frame = pygame.image.load(frame_background_path)
        frame = pygame.transform.scale(raw_frame.convert_alpha(), frame_size)

        frame_background_hover_path = (
            os.path.abspath(frame_background_hover_path)
            if frame_background_hover_path
            else default_sprites["button_background"]["active"]
        )
        raw_frame_hover = pygame.image.load(frame_background_hover_path)
        frame_hover = pygame.transform.scale(
            raw_frame_hover.convert_alpha(), frame_size
        )

        if image_path:
            image = pygame.transform.scale(
                pygame.image.load(image_path),
                (frame_size[0] - padding * 2, frame_size[1] - padding * 2),
            )
            frame.blit(image, (padding, padding))
            frame_hover.blit(image, (padding, padding))

        self.sprite.blit(frame, frame_position)
        self.sprite_hover.blit(frame_hover, frame_position)

    def render_sprite(
        self, background_path: str, rendered_text_lines: Sequence[pygame.Surface]
    ) -> pygame.Surface:
        """
        Compute the rendering of the image button with the given background and text lines.
        Render the text lines next to the image frame.

        Return the generated pygame Surface.

        Keyword arguments:
        background_path -- the path to the image corresponding to the sprite of the button
        rendered_text_lines -- the sequence of text lines in order that should be clipped
        on the surface
        """
        raw_sprite = pygame.image.load(background_path)
        sprite = pygame.transform.scale(raw_sprite.convert_alpha(), self.size)

        text_lines_count = len(rendered_text_lines)

        for index, rendered_text_line in enumerate(rendered_text_lines):
            sprite.blit(
                rendered_text_line,
                (
                    self.size[1],
                    (2 * index + 1) * sprite.get_height() // (2 * text_lines_count)
                    - rendered_text_line.get_height() // 2,
                ),
            )
        return sprite



"""
Defines InfoBox class, an helper to draw any kind of popup (menu or informative message).
"""
import os.path
from typing import Union, Sequence, List, Callable, Optional

import pygame

from ..configuration import default_sprites, default_fonts
from ..constants import (
    WHITE,
    CLOSE_BUTTON_MARGIN_TOP,
    MARGIN_BOX,
    DEFAULT_MARGIN_TOP,
    CLOSE_BUTTON_SIZE,
    MARGIN_LINKED_ELEMENT,
    DEFAULT_POPUP_WIDTH,
)
from .box_element import BoxElement
from .text_element import TextElement
from .button import Button
from ..types import Position


class _Row:
    def __init__(self, elements: List[BoxElement], height: int = 0):
        self.elements = elements
        self.height = height


class InfoBox:
    """
    This class is defining any kind of popup that can be found in the app.
    It can be used to represent the interface of a menu, or a simple text message.
    Some elements can be buttons, that will react to user clicks (see the button component
    for more information).


    Keyword arguments:
    title -- the title of the infoBox
    element_grid -- a grid containing the components that should be rendered by the infoBox
    width -- the width of the infoBox, DEFAULT_POPUP_WIDTH will be assigned if none is given
    element_linked -- the pygame Rect of the element linked to this infoBox if any,
    the infoBox will be displayed beside the element if provided
    has_close_button -- a boolean indicating whether a close button should be added
    at the bottom or not
    title_color -- the color of the title
    background_path -- the path corresponding to the image that will be the sprite of
    the infoBox
    close_button_sprite -- the path to the image corresponding to the sprite of the close button if there should be one
    close_button_sprite_hover -- the path to the image corresponding to the sprite of the close button when it is hovered if there should be one
    visible_on_background -- a boolean indicating whether the popup is visible on background or not
    has_vertical_separator -- a boolean indicating if there should be a line splitting the infoBox in two
    at middle width or not

    Attributes:
    title -- the title of the infoBox
    element_linked -- the pygame Rect of the element linked to this infoBox if there is one
    has_close_button -- whether the infoBox has a close button or not
    title_color -- the color of the title
    element_grid -- the grid containing the components that should be rendered by the infoBox
    buttons -- the sequence of buttons of the infoBox, including the close button if present
    sprite -- the pygame Surface corresponding to the sprite of the infoBox
    visible_on_background -- whether the popup is visible on background or not
    """

    def __init__(
            self,
            title: str,
            element_grid: List[List[BoxElement]],
            width: int = DEFAULT_POPUP_WIDTH,
            element_linked: pygame.Rect = None,
            has_close_button: bool = True,
            title_color: pygame.Color = WHITE,
            background_path: str = None,
            close_button_background_path: str = None,
            close_button_background_hover_path: str = None,
            visible_on_background: bool = True,
            has_vertical_separator: bool = False,
            kind: str = ""
    ) -> None:
        self.title: str = title
        self.element_linked: pygame.Rect = element_linked
        self.has_close_button: bool = has_close_button
        self.title_color: pygame.Color = title_color
        self.element_grid: List[List[BoxElement]] = element_grid
        self.__separator: dict[str, Union[bool, int]] = {
            "display": has_vertical_separator,
            "vertical_position": 0,
            "height": 0,
        }
        background_path = (
            os.path.abspath(background_path)
            if background_path
            else default_sprites["info_box_background"]
        )
        self.sprite: pygame.Surface = pygame.image.load(background_path)
        self.__close_button_background_path: str = close_button_background_path
        self.__close_button_background_hover_path: str = (
            close_button_background_hover_path
        )
        self.__elements: List[_Row] = self.init_elements(width)
        self.buttons: Sequence[Button] = []
        self.__size: tuple[int, int] = (width, 0)
        self.__position: Position = pygame.Vector2(0, 0)
        self.visible_on_background: bool = visible_on_background
        self.type: str = kind

    def init_render(
            self, screen: pygame.Surface, close_button_callback: Callable = None
    ) -> None:
        """
        Initialize the rendering of the popup.
        Compute it size and its position according to the given screen.
        Determine the position of each component.

        Keyword arguments:
        screen -- the screen on which the popup is
        close_button_callback -- the callback that should be executed when clicking on
        the close button if there is any
        """
        if self.has_close_button:
            self.__elements[-1].elements[0].callback = close_button_callback
        self.__resize_elements()
        height: int = self.__determine_height()
        self.__size = (self.__size[0], height)
        self.__position = self.determine_position(screen)
        if self.__position:
            self.determine_elements_position()
        self.buttons: Sequence[Button] = self.find_buttons()
        self.sprite = pygame.transform.scale(self.sprite.convert_alpha(), self.__size)
        self.__separator["height"] += height

    def init_elements(self, width: int) -> List[_Row]:
        """
        Initialize the graphical elements associated to the formal data that the infoBox should
        represent.

        Return the elements in a 2D structure to know the relative position of each element.

        Keyword arguments:
        width -- the width of the infoBox
        """
        elements: List[_Row] = [_Row(element_line) for element_line in self.element_grid]
        title = TextElement(
            self.title,
            pygame.Vector2(0, 0),
            default_fonts["info_box_title"],
            (20, 0, 20, 0),
            self.title_color,
        )
        self.__separator["vertical_position"] += title.get_height()
        elements.insert(0, _Row([title]))
        if self.has_close_button:
            elements.append(
                _Row([
                        Button(
                            size=CLOSE_BUTTON_SIZE,
                            title="Close",
                            background_path=self.__close_button_background_path,
                            background_hover_path=self.__close_button_background_hover_path,
                            margin=(CLOSE_BUTTON_MARGIN_TOP, 0, 0, 0),
                        )
                    ]
                )
            )
        return elements

    def __determine_height(self) -> int:
        """
        Compute the total height of the infoBox, defined according
        to the height of each element in it.
        Return the computed height.

        Keyword arguments:
        close_button_action -- the callback to run when pressing the close button
        if there should be one
        """
        # Margin to be add at begin and at end
        height: int = MARGIN_BOX * 2
        self.__separator["height"] -= height
        self.__separator["vertical_position"] += height

        for row in self.__elements:
            max_height: int = 0
            for element in row.elements:
                el_height = element.get_height() + DEFAULT_MARGIN_TOP
                if el_height > max_height:
                    max_height = el_height
            height += max_height
            row.height = max_height

        if self.has_close_button:
            self.__separator["height"] -= self.__elements[-1].height

        return height

    def __resize_elements(self) -> None:
        """
        Resize elements according to the current width of the infoBox
        """
        for row in self.__elements:
            for element in row.elements:
                if isinstance(element, TextElement):
                    element.content = element.verify_rendered_text_size(
                        element.content,
                        element.text,
                        self.__size[0]
                        - element.get_margin_left()
                        - element.get_margin_right(),
                    )
                    element.size = element.content.get_size()

    def determine_position(self, screen: pygame.Surface) -> Optional[Position]:
        """
        Compute the position of the infoBox to be beside the linked element.
        If no element is linked to the infoBox, the position will be determine at display time
        according to the screen.
        Return the computed position.
        """
        if self.element_linked:
            position: Position = pygame.Vector2(
                self.element_linked.x
                + self.element_linked.width
                + MARGIN_LINKED_ELEMENT,
                self.element_linked.y
                + self.element_linked.height // 2
                - self.__size[1] // 2,
            )
            if position.y < 0:
                position.y = 0
            elif position.y + self.__size[1] > screen.get_height():
                position.y = screen.get_height() - self.__size[1]
            if position.x + self.__size[0] > screen.get_width():
                position.x = self.element_linked.x - self.__size[0]
            return position
        return None

    def find_buttons(self) -> Sequence[Button]:
        """
        Search in all elements for buttons.
        Return the sequence of buttons.
        """
        buttons: List[Button] = []
        for row in self.__elements:
            for element in row.elements:
                if isinstance(element, Button):
                    buttons.append(element)
        return buttons

    def determine_elements_position(self):
        """
        Compute the position of each element and update it if needed.
        """
        y_coordinate: int = self.__position[1] + MARGIN_BOX
        # Memorize mouse position in case it is over a button
        mouse_pos = pygame.mouse.get_pos()
        # A row begins by a value identifying its height, followed by its elements
        for row in self.__elements:
            nb_elements = len(row.elements)
            i = 1
            for element in row.elements:
                base_x = self.__position.x + (self.__size[0] // (2 * nb_elements)) * i
                x_coordinate = base_x - element.get_width() // 2
                element.position = pygame.Vector2(
                    x_coordinate,
                    y_coordinate + element.get_margin_top(),
                )
                if isinstance(element, Button):
                    element.set_hover(element.get_rect().collidepoint(mouse_pos))
                i += 2
            y_coordinate += row.height

    def display(self, screen: pygame.Surface) -> None:
        """
        Display the infoBox and all its elements.

        Keyword arguments:
        screen -- the screen on which the displaying should be done
        """
        if self.__position:
            screen.blit(self.sprite, self.__position)
        else:
            win_size = screen.get_size()
            self.__position = pygame.Vector2(
                win_size[0] // 2 - self.__size[0] // 2,
                win_size[1] // 2 - self.__size[1] // 2,
            )
            screen.blit(self.sprite, self.__position)
            self.determine_elements_position()

        for row in self.__elements:
            for element in row.elements:
                element.display(screen)

        if self.__separator["display"]:
            pygame.draw.line(
                screen,
                WHITE,
                (
                    self.__position.x + self.__size[0] / 2,
                    self.__position.y + self.__separator["vertical_position"],
                ),
                (
                    self.__position.x + self.__size[0] / 2,
                    self.__position.y + self.__separator["height"],
                ),
                2,
            )

    def motion(self, position: Position) -> None:
        """
        Handle the triggering of a motion event.
        Test if the mouse entered in a button or quited one.

        Keyword arguments:
        position -- the position of the mouse
        """
        for button in self.buttons:
            mouse_is_on_button: bool = button.get_rect().collidepoint(position)
            button.set_hover(mouse_is_on_button and not button.disabled)

    def click(self, position: Position) -> Callable:
        """
        Handle the triggering of a click event.
        Return the data corresponding of the action that should be done if the click was done
        on a button, else False.

        Keyword arguments:
        position -- the position of the mouse
        """
        for button in self.buttons:
            if button.get_rect().collidepoint(position):
                return button.action_triggered()
        # Return a " do nothing " callable when clicking on empty space
        return lambda: None



"""
Defines the initialization process of the package through the function init.
"""

from . import fonts


def init() -> None:
    """
    Initialize the package modules.
    Must be called before all other calls and after the initialization of pygame.
    """
    fonts.init()



"""
Defines BoxElement class, useful for drawing elements.
All other components that should be drawn on a popup inherit from this class.
"""

from typing import Optional

import pygame

from ..types import Position, Margin


class BoxElement:
    """
    This element acts as a wrapper for a gui element.
    In fact, it adds a margin to any border of an element to have a cleaner interface.

    Keyword arguments:
    position -- the position of the box on the screen
    content -- a pygame surface that will be wrapped by the box
    margin -- a tuple containing the margins of the box,
    should be in the form "(top_margin, right_margin, bottom_margin, left_margin)"

    Attributes:
    position -- the position of the box on the screen
    content -- the element wrapped in the box
    size -- the size of the content following the format "(width, height)"
    margin -- a dict containing all the values for margins TOP, BOTTOM, LEFT and RIGHT
    """

    def __init__(
        self,
        position: Position,
        content: Optional[pygame.Surface],
        margin: Margin = (0, 0, 0, 0),
    ) -> None:
        self.position: Position = position
        self.content: pygame.Surface = content
        self.size: tuple[int, int] = (0, 0)
        if self.content:
            self.size = (self.content.get_width(), self.content.get_height())
        self.margin: dict[str, int] = {
            "TOP": margin[0],
            "BOTTOM": margin[2],
            "LEFT": margin[3],
            "RIGHT": margin[1],
        }

    def get_width(self) -> int:
        """
        Return the width of the content more the left and right margins
        """
        return self.margin["LEFT"] + self.size[0] + self.margin["RIGHT"]

    def get_height(self) -> int:
        """
        Return the height of the content more the top and bottom margins
        """
        return self.margin["TOP"] + self.size[1] + self.margin["BOTTOM"]

    def get_margin_top(self) -> int:
        """
        Return top margin
        """
        return self.margin["TOP"]

    def get_margin_bottom(self) -> int:
        """
        Return bottom margin
        """
        return self.margin["BOTTOM"]

    def get_margin_left(self) -> int:
        """
        Return left margin
        """
        return self.margin["LEFT"]

    def get_margin_right(self) -> int:
        """
        Return right margin
        """
        return self.margin["RIGHT"]

    def get_rect(self) -> pygame.Rect:
        """
        Return a pygame rect containing the position of the element and its size
        """
        return pygame.Rect(
            self.position[0], self.position[1], self.size[0], self.size[1]
        )

    def display(self, screen: pygame.Surface) -> None:
        """
        Display the content of the box, following the margins that should be added around it.

        Keyword arguments:
        screen -- the screen on which the content of the box should be drawn
        """
        screen.blit(
            self.content, (self.position[0] + self.margin["LEFT"], self.position[1])
        )



"""
Defines fonts that will be used all over the application.
The init_fonts function should be called at the beginning of the application
after pygame initialization.
"""

import pygame

from .configuration import default_fonts


def init() -> None:
    """
    Load all fonts registered in default_fonts.
    System font will be load if the keyword 'default' is present in the description provided.
    These fonts will be available in all modules by importing the fonts dictionary.
    """
    for font_name, font in default_fonts.items():
        if font["is_system_font"]:
            # Use pygame's default font
            is_bold = font["is_bold"] if "is_bold" in font else False
            default_fonts[font_name] = pygame.font.SysFont(
                "arial", font["size"], is_bold
            )
        else:
            default_fonts[font_name] = pygame.font.Font(font["name"], font["size"])



"""
Defines MenuManager class, the main class to handle the displaying of all menus on a screen and
handle the triggering of user events on the elements of the active menu
"""

from typing import Optional, List

import pygame

from .components.info_box import InfoBox
from .types import Position


class MenuManager:
    """
    This class represents a manager for the interfaces of a screen.
    Display all menu currently registered in the manager.
    Handle the opening of a new menu and the closing of the active one.
    Handle the triggering of user motion events and user click events on the active menu.

    Keyword arguments:
    screen -- the screen on which the menus should be displayed and on which the
    user events should be handled

    Attributes:
    screen -- the screen on which the menus should be displayed and on which the
    user events should be handled
    active_menu -- the current menu in the foreground, the only one that will react to user events
    background_menus -- the ordered sequence of menus that are in the background
    """

    def __init__(self, screen: pygame.Surface) -> None:
        self.screen: pygame.Surface = screen
        self.active_menu: Optional[InfoBox] = None
        self.background_menus: List[InfoBox] = []

    def open_menu(self, menu: InfoBox) -> None:
        """
        Open the given menu.
        Initialize the rendering of the menu, and set it as the new active menu.
        Previous active menu is sent to the background.

        Keyword arguments:
        menu -- the popup that should be open
        """
        menu.init_render(self.screen, close_button_callback=lambda: self.close_active_menu())
        if self.active_menu:
            self.background_menus.append(self.active_menu)
        self.active_menu = menu

    def close_active_menu(self) -> None:
        """
        Close the active menu by 'destroying' it.
        Take the next menu in the background to move it to foreground if there is any.
        """
        self.active_menu = (
            self.background_menus.pop() if len(self.background_menus) != 0 else None
        )
        if self.active_menu:
            # Trigger an irrelevant motion event to refresh the hovering of buttons on the new menu
            self.active_menu.motion(pygame.Vector2(pygame.mouse.get_pos()))

    def clear_menus(self) -> None:
        """
        Close all the menus (in foreground and in background)
        """
        self.active_menu = None
        self.background_menus.clear()

    def reduce_active_menu(self) -> None:
        """
        Move the active menu to the background.
        """
        if self.active_menu:
            self.background_menus.append(self.active_menu)
            self.active_menu = None

    def display(self) -> None:
        """
        Display all the visible menus in the background in order first, then display the active menu
        """
        for menu in self.background_menus:
            if menu.visible_on_background:
                menu.display(self.screen)
        if self.active_menu:
            self.active_menu.display(self.screen)

    def click(self, button: int, position: Position) -> None:
        """
        Handle the triggering of a click event.
        Delegate this event to the active menu if there is any and if it's a left click.

        Keyword arguments:
        button -- an integer value representing which mouse button has been pressed
        (1 for left button, 2 for middle button, 3 for right button)
        position -- the position of the mouse
        """
        if button == 1:
            if self.active_menu:
                self.active_menu.click(position)()

    def motion(self, position: Position) -> None:
        """
        Handle the triggering of a motion event.
        Delegate this event to the active menu if there is any.

        Keyword arguments:
        position -- the position of the mouse
        """
        if self.active_menu:
            self.active_menu.motion(position)



import os
from typing import Union

import pkg_resources
import pygame

resource_package = __name__

default_sprites = {
    "button_background": {
        "inactive": pkg_resources.resource_filename(
            resource_package, "/".join(("images", "default_box.png"))
        ),
        "active": pkg_resources.resource_filename(
            resource_package, "/".join(("images", "default_box_hover.png"))
        ),
    },
    "dynamic_button_background": {
        "inactive": pkg_resources.resource_filename(
            resource_package, "/".join(("images", "default_box.png"))
        ),
        "active": pkg_resources.resource_filename(
            resource_package, "/".join(("images", "default_box_hover.png"))
        ),
    },
    "info_box_background": pkg_resources.resource_filename(
        resource_package, "/".join(("images", "default_box.png"))
    ),
}

default_fonts: dict[str, Union[dict[str, any], pygame.font.Font]] = {
    "button_title": {"is_system_font": True, "size": 20, "is_bold": True},
    "dynamic_button_title": {"is_system_font": True, "size": 20, "is_bold": True},
    "text_element_content": {"is_system_font": True, "size": 20, "is_bold": True},
    "info_box_title": {"is_system_font": True, "size": 40, "is_bold": True},
}


def set_button_background(
    button_background_path: str, button_hovered_background_path: str
) -> None:
    default_sprites["button_background"]["inactive"] = os.path.abspath(
        button_background_path
    )
    default_sprites["button_background"]["active"] = os.path.abspath(
        button_hovered_background_path
    )


def set_dynamic_button_background(
    button_background_path: str, button_hovered_background_path: str
) -> None:
    default_sprites["dynamic_button_background"]["inactive"] = os.path.abspath(
        button_background_path
    )
    default_sprites["dynamic_button_background"]["active"] = os.path.abspath(
        button_hovered_background_path
    )


def set_info_box_background(info_box_background_path: str) -> None:
    default_sprites["info_box_background"] = os.path.abspath(info_box_background_path)


def set_button_title_font(font: pygame.font.Font) -> None:
    default_fonts["button_title"] = font


def set_dynamic_button_title_font(font: pygame.font.Font) -> None:
    default_fonts["dynamic_button_title"] = font


def set_text_element_font(font: pygame.font.Font) -> None:
    default_fonts["text_element_content"] = font


def set_info_box_title_font(font: pygame.font.Font) -> None:
    default_fonts["info_box_title"] = font



"""
Defines Button class, a BoxElement able to react to user actions.
"""
import os.path
from enum import Enum
from typing import Union, Callable, Sequence

import pygame

from ..configuration import default_sprites, default_fonts
from .box_element import BoxElement
from ..constants import WHITE, BUTTON_SIZE
from ..types import Position, Margin


class Button(BoxElement):
    """
    This class is representing any kind of button that could be seen on an interface.
    A button is receptive to user clicks and returns an id corresponding to a method,
    it may have specific arguments too.
    Mouse motion is also handled: the button appearance can change according to current focus.

    Keyword arguments:
    callback -- the reference to the function that should be call after a click
    size -- the size of the button following the format "(width, height)"
    title -- the text that should be displayed at the center of the element
    position -- the position of the element on the screen
    background_path -- the path to the image corresponding to the sprite of the element
    background_hover_path -- the path to the image corresponding to the sprite of the element
    when it has the focus
    no_background -- specify whether a background should be present or not
    margin -- a tuple containing the margins of the box,
    should be in the form "(top_margin, right_margin, bottom_margin, left_margin)"
    disabled -- a boolean indicating if it is not possible to interact with the button
    font -- the font that should be used to render the text content
    text_color -- the color of the text content
    font_hover -- the font that should be used to render the text content when the mouse is over
    the button
    text_hover_color -- the color of the text content when the mouse is over the button
    complementary_text_lines -- the other text lines that should be displayed in addition of
    the title

    Attributes:
    callback -- the reference to the function that should be call after a click
    sprite -- the pygame Surface corresponding to the sprite of the element
    sprite_hover -- the pygame Surface corresponding to the sprite of the element
    when it has the focus
    """

    def __init__(
        self,
        callback: Callable = lambda: None,
        size: tuple[int, int] = BUTTON_SIZE,
        title: str = "",
        position: Position = pygame.Vector2(0, 0),
        background_path: str = None,
        background_hover_path: str = None,
        no_background: bool = False,
        margin: Margin = (0, 0, 0, 0),
        disabled: bool = False,
        font: pygame.font.Font = None,
        text_color: tuple[int, int, int] = WHITE,
        font_hover: pygame.font.Font = None,
        text_hover_color: tuple[int, int, int] = WHITE,
        complementary_text_lines: Sequence[str] = None,
    ) -> None:
        super().__init__(position, None, margin)
        self.callback: Union[Enum, Callable] = callback
        self.size: tuple[int, int] = size

        if complementary_text_lines is None:
            complementary_text_lines = []
        text_lines = [title] + complementary_text_lines

        if not font:
            font = default_fonts["button_title"]
        rendered_text_lines = Button.render_text_lines(text_lines, text_color, font)
        if no_background:
            self.size = (rendered_text_lines[0].get_width(), rendered_text_lines[0].get_height() * len(rendered_text_lines))

        if no_background:
            background_path = None
        elif background_path:
            background_path = os.path.abspath(background_path)
        else:
            background_path = default_sprites["button_background"]["inactive"]
        self.sprite = self.render_sprite(background_path, rendered_text_lines)

        if not font_hover:
            font_hover = font
        rendered_text_lines_hover = Button.render_text_lines(
            text_lines, text_hover_color, font_hover
        )

        if no_background:
            background_hover_path = None
        elif background_hover_path:
            background_hover_path = os.path.abspath(background_hover_path)
        else:
            background_hover_path = default_sprites["button_background"]["active"]
        self.sprite_hover = self.render_sprite(
            background_hover_path, rendered_text_lines_hover
        )

        self.content = self.sprite
        self.disabled = disabled

    @staticmethod
    def render_text_lines(
        text_lines: Sequence[str],
        text_color: tuple[int, int, int],
        font: pygame.font.Font,
    ) -> Sequence[pygame.Surface]:
        """
        Compute the rendering of the given text.

        Return the rendered text lines.

        Keyword arguments:
        text_lines -- the sequence in order of text lines to be rendered
        text_color -- the color of the text
        font -- the font that should be used to render the text
        """
        return [font.render(text_line, True, text_color) for text_line in text_lines]

    def render_sprite(
        self, background_path: str, rendered_text_lines: Sequence[pygame.Surface]
    ) -> pygame.Surface:
        """
        Compute the rendering of the button with the given background and text lines.
        If no background is provided, render the text on an empty surface.

        Return the generated pygame Surface.

        Keyword arguments:
        background_path -- the path to the image corresponding to the sprite of the button
        rendered_text_lines -- the sequence of text lines in order that should be clipped
        on the surface
        """
        raw_sprite = pygame.image.load(background_path) if background_path else pygame.Surface((0, 0))
        sprite = pygame.transform.scale(raw_sprite.convert_alpha(), self.size)
        text_lines_count = len(rendered_text_lines)

        for index, rendered_text_line in enumerate(rendered_text_lines):
            sprite.blit(
                rendered_text_line,
                (
                    sprite.get_width() // 2 - rendered_text_line.get_width() // 2,
                    (2 * index + 1) * sprite.get_height() // (2 * text_lines_count)
                    - rendered_text_line.get_height() // 2,
                ),
            )
        return sprite

    def set_hover(self, is_mouse_hover: bool) -> None:
        """
        Change the current sprite between sprite or sprite_hover
        depending on whether the mouse is over the element or not.

        Keyword arguments:
        is_mouse_hover -- a boolean value indicating if the mouse is over the element or not
        """
        self.content = self.sprite_hover if is_mouse_hover else self.sprite

    def action_triggered(self) -> Callable:
        """
        Method that should be called after a click.

        Return the callback that should be executed.
        Return a callback doing nothing if the button is disabled.
        """
        if self.disabled:
            return lambda: None
        return self.callback



"""
Module to keep all the constants commonly used by other scripts.

Can be imported from anywhere.
"""

import pygame

# Colors
WHITE = pygame.Color("white")
BLACK = pygame.Color("black")
MIDNIGHT_BLUE = pygame.Color("midnightblue")

# Display parameters
CLOSE_BUTTON_MARGIN_TOP = 20
DEFAULT_MARGIN_TOP = 10
MARGIN_BOX = 20
MARGIN_LINKED_ELEMENT = 20
DEFAULT_POPUP_WIDTH = 400

# Standard menus and buttons sizes
BUTTON_SIZE = (200, 60)
IMAGE_BUTTON_SIZE = (250, 60)
CLOSE_BUTTON_SIZE = (150, 50)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pygame-popup",
    version="0.4.4",
    author="Grimmys",
    author_email="grimmys.programming@gmail.com",
    description="A popup manager for pygame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Grimmys/pygame_popup_manager",
    project_urls={
        "Issue Tracker": "https://github.com/Grimmys/pygame_popup_manager/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    package_data={"": ["images/*.png"]},
    python_requires=">=3.0",
    install_requires=["pygame~=2.0.0"],
)



"""
Defines TextElement class, a BoxElement permitting to easily draw text and let it
be centered on an popup.
"""

import pygame
from pygame.constants import SRCALPHA

from ..configuration import default_fonts
from ..constants import WHITE
from .box_element import BoxElement
from ..types import Position, Margin


class TextElement(BoxElement):
    """
    This class is representing a paragraph of text, displayed on a popup and horizontally
    centered according to its position.

    Keyword arguments:
    text -- the text that should be rendered
    position -- the position of the text on the screen
    font -- the font that should be used to render the text
    margin -- a tuple containing the margins of the box,
    should be in the form "(top_margin, right_margin, bottom_margin, left_margin)"
    text_color -- the color of the rendered text
    """

    def __init__(
        self,
        text: str,
        position: Position = pygame.Vector2(0, 0),
        font: pygame.font.Font = None,
        margin: Margin = (0, 0, 0, 0),
        text_color: pygame.Color = WHITE,
    ) -> None:
        if not font:
            font = default_fonts["text_element_content"]
        self.font = font
        self.text = text
        self.text_color = text_color
        rendered_text: pygame.Surface = font.render(text, True, text_color)
        super().__init__(position, rendered_text, margin)

    def verify_rendered_text_size(
        self, rendered_text: pygame.Surface, text: str, container_width: int
    ) -> pygame.Surface:
        """
        Split a given text in multiple lines until it could fit properly in its container

        Return the final rendered text

        Keyword arguments:
        rendered_text -- the current rendering of the text, to check if it fits in the container
        text -- the text that would be split if necessary
        container_width -- the width of the container
        """
        final_render = rendered_text

        if final_render.get_width() + 20 > container_width:
            first_part, second_part = TextElement.divide_text(text)
            first_part_render = self.font.render(first_part, True, self.text_color)
            first_part_render = self.verify_rendered_text_size(
                first_part_render, first_part, container_width
            )
            second_part_render = self.font.render(second_part, True, self.text_color)
            second_part_render = self.verify_rendered_text_size(
                second_part_render, second_part, container_width
            )
            final_render = pygame.Surface(
                (
                    container_width,
                    first_part_render.get_height() + second_part_render.get_height(),
                ),
                SRCALPHA,
            )
            first_part_x = (
                final_render.get_width() // 2 - first_part_render.get_width() // 2
            )
            final_render.blit(first_part_render, (first_part_x, 0))
            second_part_x = (
                final_render.get_width() // 2 - second_part_render.get_width() // 2
            )
            final_render.blit(
                second_part_render, (second_part_x, first_part_render.get_height())
            )
        return final_render

    @staticmethod
    def divide_text(text: str) -> tuple[str, str]:
        """
        Divide a text in two parts of a similar size, avoiding to cut a word in two.

        Return the split text in two different strings.

        Keyword argument:
        text -- the text that should be divided
        """
        separation_index: int = TextElement.get_middle_text(text)
        return text[:separation_index], text[separation_index:]

    @staticmethod
    def get_middle_text(text: str) -> int:
        """
        Returns the index of the first whitespace character that is after the middle of the provided
        string.
        If there is no whitespace character after the middle of the string, -1 is returned.

        Keyword attributes:
        text -- the string that should be used to make the computation
        """
        absolute_middle = len(text) // 2
        for i in range(absolute_middle, len(text)):
            if text[i] == " ":
                return i
        return -1



"""
Defines DynamicButton class, a special Button iterating through a list of values
on each click.
Generally use as a parameter button.
"""

from typing import Sequence, Callable

import pygame

from ..configuration import default_fonts
from ..constants import WHITE, BUTTON_SIZE
from ..types import Position, Margin
from .button import Button


class DynamicButton(Button):
    """
    This class is representing a special button with an inner value changing after each click.
    A DynamicButton has a sequence of values given at initialization, and a initial value.
    The sequence will be iterated to determine the next inner value after a click.
    This fluctuating value is the one that will be send has the first argument of
    the method called on click, and a different label will be displayed on the button for
    each different value of the sequence.

    Keyword arguments:
    callback -- the reference to the function that should be call after a click
    size -- the size of the button following the format "(width, height)"
    values -- the sequence of values that will be iterated to determine the next inner value
    current_value_index -- the index of the initial value of the button
    base_title -- the common prefix of all the different labels
    (it could be the name of the dynamic button in a way)
    position -- the position of the element on the screen
    background_path -- the path to the image corresponding to the sprite of the element
    background_hover_path -- the path to the image corresponding to the sprite of the element
    when it has the focus
    no_background -- specify whether a background should be present or not
    margin -- a tuple containing the margins of the box,
    should be in the form "(top_margin, right_margin, bottom_margin, left_margin)"
    disabled -- a boolean indicating if it is not possible to interact with the button

    Attributes:
    values -- the sequence of values that will be iterated to determine the next inner value
    current_value_index -- the index of the current value of the button
    base_title -- the common prefix of all the different labels
    (it could be the name of the dynamic button in a way)
    """

    # TODO: it should be possible to provide a specific font / font hover
    def __init__(
        self,
        callback: Callable,
        values: Sequence[any],
        current_value_index: int,
        base_title: str,
        size: tuple[int, int] = BUTTON_SIZE,
        position: Position = pygame.Vector2(0, 0),
        background_path: str = None,
        background_hover_path: str = None,
        no_background: bool = False,
        margin: Margin = (0, 0, 0, 0),
        disabled: bool = False,
    ) -> None:
        # TODO: default background for dynamic button should be used instead of letting the ascendant init takes
        #  the default one for generic button
        super().__init__(
            callback,
            size,
            "",
            position,
            background_path,
            background_hover_path,
            no_background,
            margin,
            disabled,
        )
        self.values: Sequence[any] = values
        self.current_value_index: int = current_value_index
        self.base_title: str = base_title
        self.__base_sprite: pygame.Surface = self.sprite
        self.__base_sprite_hover: pygame.Surface = self.sprite_hover
        self.__update_sprite()

    def __update_sprite(self) -> None:
        """
        Update the render of the button to display the updated dynamic value.
        Should be called after the current value changed.
        """
        rendered_name: pygame.Surface = default_fonts["dynamic_button_title"].render(
            f'{self.base_title} {self.values[self.current_value_index]["label"]}',
            True,
            WHITE,
        )

        temporary_sprite: pygame.Surface = self.__base_sprite.copy()
        temporary_sprite.blit(
            rendered_name,
            (
                temporary_sprite.get_width() // 2 - rendered_name.get_width() // 2,
                temporary_sprite.get_height() // 2 - rendered_name.get_height() // 2,
            ),
        )
        self.sprite = temporary_sprite

        temporary_sprite_hover: pygame.Surface = self.__base_sprite_hover.copy()
        temporary_sprite_hover.blit(
            rendered_name,
            (
                temporary_sprite_hover.get_width() // 2
                - rendered_name.get_width() // 2,
                temporary_sprite_hover.get_height() // 2
                - rendered_name.get_height() // 2,
            ),
        )
        self.sprite_hover = temporary_sprite_hover

        # Force display update
        self.set_hover(True)

    def action_triggered(self) -> Callable:
        """
        Method that should be called after a click.
        Change the current value of the button to the next one in the sequence of values.
        If the end of the sequence is reach, the iteration restarts at the first value.

        Return a lambda containing the function that should be called with the current value of the
        dynamic button as an argument.
        """
        # Search for next value
        self.current_value_index += 1
        if self.current_value_index == len(self.values):
            self.current_value_index = 0
        self.__update_sprite()
        current_value = self.values[self.current_value_index]["value"]
        function_to_call: Callable = super().action_triggered()
        return lambda: function_to_call(current_value)
